# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
import mock
import pandas as pd

from copycat import ad_copy_evaluator
from copycat import ad_copy_generator
from copycat import google_ads
from copycat import testing_utils


class AdCopyEvaluatorTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.embedding_model_patcher = testing_utils.PatchEmbeddingsModel()
    self.embedding_model_patcher.start()

    training_data = pd.DataFrame({
        "headlines": [[f"train headline {i}"] for i in range(20)],
        "descriptions": [[f"train description {i}"] for i in range(20)],
        "keywords": [f"keyword {i}a, keyword {i}b" for i in range(20)],
    })

    self.ad_copy_vectorstore = (
        ad_copy_generator.AdCopyVectorstore.create_from_pandas(
            training_data=training_data,
            embedding_model_name="text-embedding-004",
            dimensionality=256,
            max_initial_ads=100,
            max_exemplar_ads=20,
            affinity_preference=None,
            embeddings_batch_size=10,
            exemplar_selection_method="random",
        )
    )

    self.ad_format = google_ads.GoogleAdFormat(
        name="test_format",
        max_headlines=15,
        min_headlines=3,
        max_descriptions=4,
        min_descriptions=2,
        max_headline_length=30,
        max_description_length=90,
    )

  def tearDown(self):
    super().tearDown()
    self.embedding_model_patcher.stop()

  def test_similarity_metrics_are_calculated_correctly(self):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(
        self.ad_format,
        ad_copy_vectorstore=self.ad_copy_vectorstore,
    )

    generated_ad = google_ads.GoogleAd(
        headlines=["generated headline"], descriptions=["generated description"]
    )

    actual_metrics = evaluator.calculate_similarity_metrics_batch(
        ad_copies=[generated_ad],
        keywords=["keyword 1, keyword 2"],
    )[0]

    expected_metrics = dict(
        style_similarity=0.5366996856066526,
        keyword_similarity=0.48784862680427793,
    )
    self.assertDictEqual(actual_metrics, expected_metrics)

  @parameterized.named_parameters([
      {
          "testcase_name": "valid",
          "n_headlines": 15,
          "expected_response": True,
      },
      {
          "testcase_name": "too many",
          "n_headlines": 16,
          "expected_response": False,
      },
      {
          "testcase_name": "too few",
          "n_headlines": 2,
          "expected_response": False,
      },
  ])
  def test_has_valid_number_of_headlines(self, n_headlines, expected_response):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(self.ad_format)

    google_ad = google_ads.GoogleAd(
        headlines=["Headline 1"] * n_headlines,
        descriptions=["Description 1.", "Description 2."],
    )

    self.assertEqual(
        evaluator.has_valid_number_of_headlines(google_ad), expected_response
    )

  @parameterized.named_parameters([
      {
          "testcase_name": "valid",
          "n_descriptions": 4,
          "expected_response": True,
      },
      {
          "testcase_name": "too many",
          "n_descriptions": 5,
          "expected_response": False,
      },
      {
          "testcase_name": "too few",
          "n_descriptions": 1,
          "expected_response": False,
      },
  ])
  def test_has_valid_number_of_descriptions(
      self, n_descriptions, expected_response
  ):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(self.ad_format)

    google_ad = google_ads.GoogleAd(
        headlines=["Headline 1"],
        descriptions=["Description 1."] * n_descriptions,
    )

    self.assertEqual(
        evaluator.has_valid_number_of_descriptions(google_ad), expected_response
    )

  @parameterized.named_parameters([
      {
          "testcase_name": "valid",
          "headline": "a" * 30,
          "expected_response": True,
      },
      {
          "testcase_name": "too long",
          "headline": "a" * 31,
          "expected_response": False,
      },
      {
          "testcase_name": "valid dki",
          "headline": "Valid DKI {KeyWord:my keyword} 123",
          "expected_response": True,
      },
      {
          "testcase_name": "invalid dki",
          "headline": "Invalid DKI {KeyWord:my keyword} too long!",
          "expected_response": False,
      },
  ])
  def test_has_valid_headline_lengths(self, headline, expected_response):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(self.ad_format)

    google_ad = google_ads.GoogleAd(
        headlines=[headline, "headline 2"],
        descriptions=["Description 1."],
    )

    self.assertEqual(
        evaluator.has_valid_headline_lengths(google_ad), expected_response
    )

  @parameterized.named_parameters([
      {
          "testcase_name": "valid",
          "description": "a" * 90,
          "expected_response": True,
      },
      {
          "testcase_name": "too long",
          "description": "a" * 91,
          "expected_response": False,
      },
      {
          "testcase_name": "valid dki",
          "description": "Valid DKI {KeyWord:my keyword} " + "a" * 63,
          "expected_response": True,
      },
      {
          "testcase_name": "invalid dki",
          "description": "Invalid DKI {KeyWord:my keyword}" + "a" * 70,
          "expected_response": False,
      },
  ])
  def test_has_valid_description_lengths(self, description, expected_response):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(self.ad_format)

    google_ad = google_ads.GoogleAd(
        headlines=["Headline 1", "headline 2"],
        descriptions=[description, "description 2"],
    )

    self.assertEqual(
        evaluator.has_valid_description_lengths(google_ad), expected_response
    )

  @parameterized.parameters(
      "has_valid_number_of_headlines",
      "has_valid_number_of_descriptions",
      "has_valid_headline_lengths",
      "has_valid_description_lengths",
  )
  def test_is_valid_fails_if_any_check_fails(self, failing_check):
    with mock.patch.object(
        ad_copy_evaluator.AdCopyEvaluator,
        failing_check,
        return_value=False,
    ):
      evaluator = ad_copy_evaluator.AdCopyEvaluator(self.ad_format)
      google_ad = google_ads.GoogleAd(
          headlines=["Headline 1", "Headline 2"],
          descriptions=["Description 1.", "Description 2."],
      )
      self.assertFalse(evaluator.is_valid(google_ad))

  @parameterized.named_parameters([
      {
          "testcase_name": "complete",
          "n_headlines": 15,
          "n_descriptions": 4,
          "expected_response": True,
      },
      {
          "testcase_name": "too few headlines",
          "n_headlines": 14,
          "n_descriptions": 4,
          "expected_response": False,
      },
      {
          "testcase_name": "too few descriptions",
          "n_headlines": 15,
          "n_descriptions": 3,
          "expected_response": False,
      },
      {
          "testcase_name": "too many headlines",
          "n_headlines": 16,
          "n_descriptions": 4,
          "expected_response": False,
      },
      {
          "testcase_name": "too many descriptions",
          "n_headlines": 15,
          "n_descriptions": 5,
          "expected_response": False,
      },
  ])
  def test_is_complete_checks_if_headlines_and_descriptions_are_full(
      self, n_headlines, n_descriptions, expected_response
  ):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(self.ad_format)
    google_ad = google_ads.GoogleAd(
        headlines=list(map(str, range(n_headlines))),
        descriptions=list(map(str, range(n_descriptions))),
    )

    self.assertEqual(evaluator.is_complete(google_ad), expected_response)

  @parameterized.named_parameters([
      {
          "testcase_name": "enough headlines and descriptions",
          "n_headlines": 3,
          "n_descriptions": 2,
          "expected_response": False,
      },
      {
          "testcase_name": "too few headlines",
          "n_headlines": 2,
          "n_descriptions": 2,
          "expected_response": True,
      },
      {
          "testcase_name": "too few descriptions",
          "n_headlines": 3,
          "n_descriptions": 1,
          "expected_response": True,
      },
      {
          "testcase_name": "too few both",
          "n_headlines": 2,
          "n_descriptions": 1,
          "expected_response": True,
      },
  ])
  def test_is_underpopulated_checks_if_too_few_headlines_or_descriptions(
      self, n_headlines, n_descriptions, expected_response
  ):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(self.ad_format)

    google_ad = google_ads.GoogleAd(
        headlines=list(map(str, range(n_headlines))),
        descriptions=list(map(str, range(n_descriptions))),
    )

    self.assertEqual(evaluator.is_underpopulated(google_ad), expected_response)

  @parameterized.named_parameters([
      {
          "testcase_name": "has headlines and descriptions",
          "n_headlines": 1,
          "n_descriptions": 1,
          "expected_response": False,
      },
      {
          "testcase_name": "no headlines",
          "n_headlines": 0,
          "n_descriptions": 1,
          "expected_response": False,
      },
      {
          "testcase_name": "no descriptions",
          "n_headlines": 1,
          "n_descriptions": 0,
          "expected_response": False,
      },
      {
          "testcase_name": "is empty",
          "n_headlines": 0,
          "n_descriptions": 0,
          "expected_response": True,
      },
  ])
  def test_is_empty_checks_if_zero_headlines_and_descriptions(
      self, n_headlines, n_descriptions, expected_response
  ):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(self.ad_format)

    google_ad = google_ads.GoogleAd(
        headlines=list(map(str, range(n_headlines))),
        descriptions=list(map(str, range(n_descriptions))),
    )

    self.assertEqual(evaluator.is_empty(google_ad), expected_response)

  @parameterized.named_parameters(
      dict(
          testcase_name="not memorised",
          generated_headlines=["new headline", "train headline 1"],
          generated_descriptions=["new description", "train description 1"],
          allow_memorised_headlines=False,
          allow_memorised_descriptions=False,
          expected_headlines_are_memorised=False,
          expected_descriptions_are_memorised=False,
          expected_errors=[],
      ),
      dict(
          testcase_name="headline memorised but allowed",
          generated_headlines=["train headline 1"],
          generated_descriptions=["new description", "train description 1"],
          allow_memorised_headlines=True,
          allow_memorised_descriptions=False,
          expected_headlines_are_memorised=True,
          expected_descriptions_are_memorised=False,
          expected_errors=[],
      ),
      dict(
          testcase_name="headline memorised but not allowed",
          generated_headlines=["train headline 1"],
          generated_descriptions=["new description", "train description 1"],
          allow_memorised_headlines=False,
          allow_memorised_descriptions=False,
          expected_headlines_are_memorised=True,
          expected_descriptions_are_memorised=False,
          expected_errors=[
              "All headlines are memorised from the training data."
          ],
      ),
      dict(
          testcase_name="description memorised but allowed",
          generated_headlines=["new headline", "train headline 1"],
          generated_descriptions=["train description 1"],
          allow_memorised_headlines=False,
          allow_memorised_descriptions=True,
          expected_headlines_are_memorised=False,
          expected_descriptions_are_memorised=True,
          expected_errors=[],
      ),
      dict(
          testcase_name="description memorised but not allowed",
          generated_headlines=["new headline", "train headline 1"],
          generated_descriptions=["train description 1"],
          allow_memorised_headlines=False,
          allow_memorised_descriptions=False,
          expected_headlines_are_memorised=False,
          expected_descriptions_are_memorised=True,
          expected_errors=[
              "All descriptions are memorised from the training data."
          ],
      ),
      dict(
          testcase_name="both memorised but allowed",
          generated_headlines=["train headline 1"],
          generated_descriptions=["train description 1"],
          allow_memorised_headlines=True,
          allow_memorised_descriptions=True,
          expected_headlines_are_memorised=True,
          expected_descriptions_are_memorised=True,
          expected_errors=[],
      ),
      dict(
          testcase_name="both memorised but not allowed",
          generated_headlines=["train headline 1"],
          generated_descriptions=["train description 1"],
          allow_memorised_headlines=False,
          allow_memorised_descriptions=False,
          expected_headlines_are_memorised=True,
          expected_descriptions_are_memorised=True,
          expected_errors=[
              "All headlines are memorised from the training data.",
              "All descriptions are memorised from the training data.",
          ],
      ),
  )
  def test_check_ad_copy_memorisation(
      self,
      generated_headlines,
      generated_descriptions,
      allow_memorised_headlines,
      allow_memorised_descriptions,
      expected_headlines_are_memorised,
      expected_descriptions_are_memorised,
      expected_errors,
  ):
    test_ad_foramt = google_ads.GoogleAdFormat(
        name="test_format",
        max_headlines=15,
        min_headlines=1,
        max_descriptions=4,
        min_descriptions=1,
        max_headline_length=30,
        max_description_length=90,
    )
    evaluator = ad_copy_evaluator.AdCopyEvaluator(
        test_ad_foramt, self.ad_copy_vectorstore
    )
    ad_copy = google_ads.GoogleAd(
        headlines=generated_headlines, descriptions=generated_descriptions
    )

    results = evaluator.evaluate_batch(
        [ad_copy],
        allow_memorised_headlines=allow_memorised_headlines,
        allow_memorised_descriptions=allow_memorised_descriptions,
    )[0]

    self.assertEqual(
        expected_headlines_are_memorised, results.headlines_are_memorised
    )
    self.assertEqual(
        expected_descriptions_are_memorised, results.descriptions_are_memorised
    )
    self.assertCountEqual(expected_errors, results.errors)

  def test_evaluate_returns_expected_results_with_vectorstore(self):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(
        self.ad_format,
        ad_copy_vectorstore=self.ad_copy_vectorstore,
    )

    ad_copy = google_ads.GoogleAd(
        headlines=["headline 1", "headline 2", "headline 3"],
        descriptions=["description 1", "description 2"],
    )

    results = evaluator.evaluate_batch(
        [ad_copy],
        allow_memorised_headlines=False,
        allow_memorised_descriptions=False,
        keywords=["keyword 1, keyword 2"],
    )[0]

    expected_results = ad_copy_evaluator.EvaluationResults(
        errors=[],
        warnings=[],
        headlines_are_memorised=False,
        descriptions_are_memorised=False,
        keyword_similarity=0.5023139051908866,
        style_similarity=0.5617324461155475,
    )
    self.assertEqual(results, expected_results)

  def test_evaluate_returns_expected_results_without_vectorstore(self):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(
        self.ad_format,
    )

    ad_copy = google_ads.GoogleAd(
        headlines=["headline 1", "headline 2", "headline 3"],
        descriptions=["description 1", "description 2"],
    )

    results = evaluator.evaluate_batch(
        [ad_copy],
        allow_memorised_headlines=False,
        allow_memorised_descriptions=False,
        keywords=["keyword 1, keyword 2"],
    )[0]

    expected_results = ad_copy_evaluator.EvaluationResults(
        errors=[],
        warnings=[],
        headlines_are_memorised=False,
        descriptions_are_memorised=False,
        keyword_similarity=None,
        style_similarity=None,
    )
    self.assertEqual(results, expected_results)

  def test_evaluate_returns_expected_results_with_vectorstore_no_keywords(self):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(
        self.ad_format,
        ad_copy_vectorstore=self.ad_copy_vectorstore,
    )

    ad_copy = google_ads.GoogleAd(
        headlines=["headline 1", "headline 2", "headline 3"],
        descriptions=["description 1", "description 2"],
    )

    results = evaluator.evaluate_batch(
        [ad_copy],
        allow_memorised_headlines=False,
        allow_memorised_descriptions=False,
    )[0]

    expected_results = ad_copy_evaluator.EvaluationResults(
        errors=[],
        warnings=[],
        headlines_are_memorised=False,
        descriptions_are_memorised=False,
        keyword_similarity=None,
        style_similarity=None,
    )
    self.assertEqual(results, expected_results)

  def test_evaluate_returns_expected_results_with_vectorstore_empty_ad(self):
    evaluator = ad_copy_evaluator.AdCopyEvaluator(
        self.ad_format,
        ad_copy_vectorstore=self.ad_copy_vectorstore,
    )

    ad_copy = google_ads.GoogleAd(headlines=[], descriptions=[])

    results = evaluator.evaluate_batch(
        [ad_copy],
        allow_memorised_headlines=False,
        allow_memorised_descriptions=False,
        keywords=["keyword 1, keyword 2"],
    )[0]

    expected_results = ad_copy_evaluator.EvaluationResults(
        errors=[
            "Invalid number of headlines for the ad format.",
            "Invalid number of descriptions for the ad format.",
        ],
        warnings=[],
        headlines_are_memorised=False,
        descriptions_are_memorised=False,
        keyword_similarity=None,
        style_similarity=None,
    )
    self.assertEqual(results, expected_results)

  @parameterized.named_parameters([
      {
          "testcase_name": "no special variables",
          "headlines": ["headline 1", "headline 2"],
          "descriptions": ["description 1", "description 2"],
          "expected_response": False,
      },
      {
          "testcase_name": "headline dki",
          "headlines": ["headline {KeyWord:my keyword}", "headline 2"],
          "descriptions": ["description 1", "description 2"],
          "expected_response": False,
      },
      {
          "testcase_name": "description dki",
          "headlines": ["headline 1", "headline 2"],
          "descriptions": ["description {KeyWord:my keyword}", "description 2"],
          "expected_response": False,
      },
      {
          "testcase_name": "headline customizer",
          "headlines": [
              "headline {CUSTOMIZER.product:my product}",
              "headline 2",
          ],
          "descriptions": ["description 1", "description 2"],
          "expected_response": False,
      },
      {
          "testcase_name": "description customizer",
          "headlines": ["headline 1", "headline 2"],
          "descriptions": [
              "description {CUSTOMIZER.product:my product}",
              "description 2",
          ],
          "expected_response": False,
      },
      {
          "testcase_name": "headline customizer no default",
          "headlines": ["headline {CUSTOMIZER.product}", "headline 2"],
          "descriptions": ["description 1", "description 2"],
          "expected_response": True,
      },
      {
          "testcase_name": "description customizer no default",
          "headlines": ["headline 1", "headline 2"],
          "descriptions": ["description {CUSTOMIZER.product}", "description 2"],
          "expected_response": True,
      },
  ])
  def test_has_unfillable_google_ads_special_variables(
      self, headlines, descriptions, expected_response
  ):
    google_ad = google_ads.GoogleAd(
        headlines=headlines,
        descriptions=descriptions,
    )
    evaluator = ad_copy_evaluator.AdCopyEvaluator(self.ad_format)
    self.assertEqual(
        evaluator.has_unfillable_google_ads_special_variables(google_ad),
        expected_response,
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Organises keywords into semantically similar ad groups."""

import abc
import logging
import numpy as np
from sklearn import cluster
from sklearn import metrics
import tqdm

ARBITRARILY_HIGH_DISTANCE_THRESHOLD = 10.0

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


class TqdmLogger:
  """File-like class redirecting tqdm progress bar to LOGGER."""

  def write(self, msg: str) -> None:
    LOGGER.info(msg.lstrip("\r"))

  def flush(self) -> None:
    pass


class KeywordClusterer(abc.ABC):
  """Abstract class for clustering keywords.

  Attributes:
    distance_threshold: The distance threshold to use for clustering - keywords
      with a distance less than this will be considered to be in the same ad
      group. This is initialized to None and then learned when fit is called.
    max_training_samples: The maximum number of samples to use for training. If
      more samples are provided, then a random subset of the samples will be
      used.
  """

  distance_threshold: float | None
  max_training_samples: int

  def __init__(self, max_training_samples: int = 5000):
    """Initialises the keyword clusterer.

    Args:
      max_training_samples: The maximum number of samples to use for training.
        If more samples are provided, then a random subset of the samples will
        be used.
    """
    self.distance_threshold = None
    self.max_training_samples = max_training_samples

  @abc.abstractmethod
  def _fit(
      self, keyword_embeddings: np.ndarray, target_ad_groups: np.ndarray
  ) -> None:
    """Fits the clusterer to the given keyword embeddings and target ad groups.

    This should learn the distance threshold such that the predicted ad groups
    produced are as similar as possible to the target ad groups. This method
    should be implemented by subclasses.

    Args:
      keyword_embeddings: The text embeddings of the keywords as a numpy array.
      target_ad_groups: The target as groups as an array of integers.
    """
    ...

  def fit(
      self, keyword_embeddings: np.ndarray, target_ad_groups: np.ndarray
  ) -> None:
    """Fits the clusterer to the given keyword embeddings and target ad groups.

    This will learn the distance threshold such that the predicted ad groups
    produced are as similar as possible to the target ad groups. It first
    downsamples the data to max_training_samples if necessary, before fitting.

    Args:
      keyword_embeddings: The text embeddings of the keywords as a numpy array.
      target_ad_groups: The target as groups as an array of integers.
    """
    if len(target_ad_groups) > self.max_training_samples:
      idx = np.random.choice(
          len(target_ad_groups), self.max_training_samples, replace=False
      )
      keyword_embeddings = keyword_embeddings[idx, :]
      target_ad_groups = target_ad_groups[idx]

    self._fit(keyword_embeddings, target_ad_groups)

  @abc.abstractmethod
  def predict(self, keyword_embeddings: np.ndarray) -> np.ndarray:
    """Predicts the ad groups for the given keyword embeddings.

    Args:
      keyword_embeddings: The keyword embeddings to use for clustering.

    Returns:
      The predicted ad group clusters as an array of integers.
    """
    ...

  def _evaluate_targets(
      self, target_ad_groups: np.ndarray, predicted_ad_groups: np.ndarray
  ) -> float:
    """Evaluates the predicted ad groups against the target ad groups.

    Args:
      target_ad_groups: The target ad groups as an array of integers.
      predicted_ad_groups: The predicted ad groups as an array of integers.

    Returns:
      The adjusted rand score between the target and predicted ad groups.
    """
    return metrics.adjusted_rand_score(target_ad_groups, predicted_ad_groups)

  def evaluate(
      self, keyword_embeddings: np.ndarray, target_ad_groups: np.ndarray
  ) -> float:
    """First clusters the provided keyword embeddings and then evaluates.

    The predicted ad groups are evaluated against the target ad groups using
    the adjusted rand score.

    Args:
      keyword_embeddings: The keyword embeddings to use for clustering.
      target_ad_groups: The target ad groups as an array of integers.

    Returns:
      The adjusted rand score between the target and predicted ad groups.
    """
    predicted_ad_groups = self.predict(keyword_embeddings)
    return self._evaluate_targets(target_ad_groups, predicted_ad_groups)


class BirchAgglomerativeKeywordClusterer(KeywordClusterer):
  """Clusters keywords using BIRCH and agglomerative clustering.

  This first learns the optimal distance threshold for agglomerative clustering
  at the fit step.

  When predicting, BIRCH clustering is performed first to improve the
  scalability of the clustering. BIRCH uses a threshold which is the learned
  distance threshold, multiplied by the birch_scale. This scaling controls how
  much of the clustering is performed by BIRCH and how much is performed by
  agglomerative clustering. If birch_scale is set to 0, then only
  agglomerative clustering is performed, if it is set to 1 then almost all of
  the clustering will be performed by BIRCH. A higher birch scale
  leads to a more scalable algorithm but usually worse performance.

  Attributes:
    birch_scale: The scale to use for BIRCH. This is a value between 0 and 1
      that controls how much BIRCH contributes to the clustering. A high value
      means BIRCH contributes more, and will be more scalable but less accurate.
  """

  def __init__(
      self, birch_scale: float = 0.5, max_training_samples: int = 5000
  ):
    """Initialises the keyword clusterer.

    Args:
      birch_scale: The scale to use for BIRCH. This is a value between 0 and 1
        that controls how much BIRCH contributes to the clustering. A high value
        means BIRCH contributes more, and will be more scalable but less
        accurate.
      max_training_samples: The maximum number of samples to use for training.
        If more samples are provided, then a random subset of the samples will
        be used.
    """
    super().__init__(max_training_samples=max_training_samples)
    self.birch_scale = birch_scale
    if self.birch_scale >= 1.0 or self.birch_scale < 0.0:
      LOGGER.error(
          "Birch scale must be >=0 and <1, but got %f", self.birch_scale
      )
      raise ValueError(
          f"Birch scale must be >=0 and <1, but got {self.birch_scale}"
      )

  def _fit(
      self, keyword_embeddings: np.ndarray, target_ad_groups: np.ndarray
  ) -> None:
    """Fits the clusterer to the given keyword embeddings and target ad groups.

    This will learn the distance threshold such that the predicted ad groups
    produced are as similar as possible to the target ad groups. It does this
    by applying Agglomerative Clustering across a variety of distance thresholds
    and then selecting the threshold that produces the highest adjusted rand
    score when comparing the predicted ad groups to the target ad groups.

    Args:
      keyword_embeddings: The text embeddings of the keywords as a numpy array.
      target_ad_groups: The target as groups as an array of integers.
    """
    LOGGER.info(
        "Fitting the distance threshold of BirchAgglomerativeKeywordClusterer."
    )

    # Construct the tree using AgglomerativeClustering with a high distance
    # threshold. This tree tells you the order in which the keywords are merged,
    # starting with the most similar and continuing recursively until they are
    # all merged into a single cluster. We will use this tree to find the
    # optimal distance threshold.
    clustering = cluster.AgglomerativeClustering(
        n_clusters=None,
        linkage="average",
        distance_threshold=ARBITRARILY_HIGH_DISTANCE_THRESHOLD,
        compute_distances=True,
        compute_full_tree=True,
    )
    clustering.fit(keyword_embeddings)

    LOGGER.info("Finished constructing the tree with AgglomerativeClustering.")

    # Iterate through the tree, calculating the rand score at each step in the
    # clustering.
    clusters = [[i] for i in range(len(target_ad_groups))]
    predicted_ad_groups = np.arange(len(target_ad_groups))
    next_cluster_id = len(target_ad_groups)

    distance = clustering.distances_
    rand_score = []

    for i, j in tqdm.tqdm(
        clustering.children_,
        desc="Fitting distance threshold",
        file=TqdmLogger(),
        mininterval=5,
    ):
      new_cluster = clusters[i] + clusters[j]

      for k in new_cluster:
        predicted_ad_groups[k] = next_cluster_id

      clusters.append(new_cluster)
      next_cluster_id += 1

      rand_score.append(
          self._evaluate_targets(target_ad_groups, predicted_ad_groups)
      )

    # The best distance is the distance with the highest score
    distance = np.asarray(distance)
    rand_score = np.asarray(rand_score)
    self.distance_threshold = distance[np.argmax(rand_score)]
    LOGGER.info(
        "Found the optimal distance threshold: %f with rand score: %f",
        self.distance_threshold,
        np.max(rand_score),
    )

  def predict(self, keyword_embeddings: np.ndarray) -> np.ndarray:
    """Predicts the ad groups for the given keyword embeddings.

    This will first perform BIRCH clustering to produce a set of initial
    clusters, and then perform agglomerative clustering to produce the final
    clusters. If birch_scale is set to 0, then no BIRCH clustering will be
    performed.

    Args:
      keyword_embeddings: The keyword embeddings to use for clustering.

    Returns:
      The predicted ad group clusters as an array of integers.

    Raises:
      ValueError: If the distance threshold has not been learned yet (meaning
        fit() has not been called).
    """
    LOGGER.info("Predicting ad groups with BirchAgglomerativeKeywordClusterer.")
    if self.distance_threshold is None:
      LOGGER.error("Must call fit before predicting ad groups.")
      raise ValueError("Must call fit before predicting ad groups.")

    agglomerative_clustering = cluster.AgglomerativeClustering(
        n_clusters=None,
        linkage="average",
        distance_threshold=self.distance_threshold,
    )
    if self.birch_scale > 0.0:
      LOGGER.info(
          "Performing BIRCH clustering with scale: %f", self.birch_scale
      )
      clustering = cluster.Birch(
          threshold=self.distance_threshold * self.birch_scale,
          n_clusters=agglomerative_clustering,
      )
    else:
      LOGGER.info("Skipping BIRCH clustering because birch scale is 0.")
      clustering = agglomerative_clustering
    return clustering.fit_predict(keyword_embeddings)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The main module for Copycat.

Contains the code to generate copycat ad copies.
"""

import dataclasses
import logging

from google.cloud import storage
from vertexai import generative_models

from copycat import ad_copy_generator

ModelName = ad_copy_generator.ModelName


LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


STYLE_PROMPT_IF_AD_REPORT_FROM_GCS = """\
In these files is an ad report for {brand_name}, containing their ads (headlines and descriptions)
that they use on Google Search Ads for the corresponding keywords. Headlines and descriptions are lists, and
Google constructs ads by combining those headlines and descriptions together into ads. Therefore the headlines and descriptions
should be sufficiently varied that Google is able to try lots of different combinations in order to find what works best.

Use the ad report to write a comprehensive style guide for this brand's ad copies that can serve as
instruction for a copywriter to write new ad copies for {brand_name} for new lists of keywords.
Ensure that you capure strong phrases, slogans and brand names of {brand_name} in the guide.

Additionally, there could be other files included regarding the brand's style that you should consider in the style guide.{additional_style_instructions}"""

STYLE_PROMPT_IF_AD_REPORT_FROM_VECTORSTORE = """\
Below is an ad report for {brand_name}, containing their ads (headlines and descriptions)
that they use on Google Search Ads for the corresponding keywords. Headlines and descriptions are lists, and
Google constructs ads by combining those headlines and descriptions together into ads. Therefore the headlines and descriptions
should be sufficiently varied that Google is able to try lots of different combinations in order to find what works best.

Use the ad report to write a comprehensive style guide for this brand's ad copies that can serve as
instruction for a copywriter to write new ad copies for {brand_name} for new lists of keywords.
Ensure that you capure strong phrases, slogans and brand names of {brand_name} in the guide.

Additionally, there could be other files included regarding the brand's style that you should consider in the style guide.{additional_style_instructions}

Ad Report:

{example_ads_json}"""


def _clean_text_newlines(text: str) -> str:
  """Cleans text newlines.

  If there is a single newline, it is removed. If there is a double newline, it
  is not removed.

  Args:
    text: The text to clean.

  Returns:
    The cleaned text.
  """
  return (
      text.replace("\n\n", "<DOUBLE_NEWLINE>")
      .replace("\n", " ")
      .replace("<DOUBLE_NEWLINE>", "\n\n")
      .replace("  ", " ")
  )


# file types that are accepted for style guide generation with Gemini
ACCEPTED_FILE_TYPES = ["application/pdf", "text/csv"]
SafetySettingsType = (
    dict[generative_models.HarmCategory, generative_models.HarmBlockThreshold]
    | list[generative_models.SafetySetting]
)


class StyleGuideGenerator:
  """A class designed to generate brand style guides using Gemini.

  This class facilitates the retrieval of files from Google Cloud Storage (GCS),
  construction of prompts incorporating these files, and the generation of style
  guides based on provided style instructions and brand information.
  """

  file_info: list[dict[str, str]] = dataclasses.field(default_factory=list)

  def __init__(self) -> None:
    """Initializes a StyleGuideGenerator instance.

    Sets the initial file_info to an empty list.
    """
    self.file_info = []

  def get_all_files(self, bucket_name: str) -> list[dict[str, str]]:
    """Retrieves all file URIs and their MIME types from a GCS bucket.

    Args:
        bucket_name: The name of the GCS bucket.

    Returns:
        A list of dictionaries, each containing information about a file.
        Each dictionary has keys: 'uri' (the GCS URI of the file) and
        'mime_type'.
    """
    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    blobs = bucket.list_blobs()

    file_info = []
    for blob in blobs:
      if blob.content_type in ACCEPTED_FILE_TYPES:
        file_data = {
            "uri": f"gs://{bucket_name}/{blob.name}",
            "mime_type": blob.content_type,
        }
        file_info.append(file_data)
    self.file_info = file_info
    LOGGER.info("Successfully retrieved %d file URI's.", len(file_info))
    return file_info

  def generate_style_guide(
      self,
      brand_name: str = "",
      ad_copy_vectorstore: ad_copy_generator.AdCopyVectorstore | None = None,
      additional_style_instructions: str = "",
      model_name: str | ModelName = ModelName.GEMINI_1_5_PRO,
      safety_settings: SafetySettingsType | None = None,
      temperature: float = 0.95,
      top_k: int = 20,
      top_p: float = 0.95,
  ) -> generative_models._generative_models.GenerationResponse:
    """Generates a style guide using the provided model and contents.

    Args:
        brand_name: The name of the brand.
        ad_copy_vectorstore: A vectorstore containing the ad copies to use for
          style guide generation. If None then the style guide will be generated
          using only the content in the files loaded from Google Cloud Storage.
        additional_style_instructions: additional style instructions.
        model_name: The name of the generative model to use.
        safety_settings: The safety settings to use for the model.
        temperature: The temperature parameter for the model (controls
          randomness).
        top_k: The top-k sampling parameter for the model.
        top_p: The top-p sampling parameter for the model.

    Returns:
        The generated style guide text.
    """
    safety_settings = safety_settings or {}

    model_name = ModelName(model_name)
    generation_config = generative_models.GenerationConfig(
        temperature=temperature,
        top_k=top_k,
        top_p=top_p,
    )

    if additional_style_instructions:
      additional_style_instructions = (
          "\n\nAlso incorporate the following style instructions into the"
          f" style guide:\n\n{additional_style_instructions}"
      )

    content = self._construct_style_prompt(
        style_prompt_params={
            "brand_name": brand_name,
            "additional_style_instructions": additional_style_instructions,
        },
        ad_copy_vectorstore=ad_copy_vectorstore,
    )
    model = generative_models.GenerativeModel(
        model_name=model_name.value, safety_settings=safety_settings
    )
    response = model.generate_content(
        contents=[content], generation_config=generation_config
    )
    if not isinstance(response, generative_models.GenerationResponse):
      LOGGER.error(
          "Response is not a GenerationResponse. Instead got: %s", response
      )
      raise RuntimeError(
          f"Response is not a GenerationResponse. Instead got: {response}"
      )
    return response

  def _construct_style_prompt(
      self,
      style_prompt_params: dict[str, str] | None = None,
      ad_copy_vectorstore: ad_copy_generator.AdCopyVectorstore | None = None,
  ) -> generative_models.Content:
    """Constructs the prompt for style guide generation.

    Args:
        style_prompt_params: The parameters to use for the style prompt.
        ad_copy_vectorstore: The ad copy vectorstore to use for the style
          prompt.

    Returns:
        The constructed prompt in Content format.
    """
    style_prompt_params = style_prompt_params or {}

    if ad_copy_vectorstore is not None:
      LOGGER.info("Using ad copy vectorstore for style guide generation.")
      style_prompt_params["example_ads_json"] = (
          ad_copy_vectorstore.ad_exemplars[
              ["keywords", "headlines", "descriptions"]
          ].to_json(orient="records", indent=1, force_ascii=False, lines=True)
      )

    if ad_copy_vectorstore is not None:
      style_prompt = STYLE_PROMPT_IF_AD_REPORT_FROM_VECTORSTORE
    else:
      style_prompt = STYLE_PROMPT_IF_AD_REPORT_FROM_GCS

    full_style_prompt = _clean_text_newlines(
        style_prompt.format(**style_prompt_params)
    )

    contents: list[generative_models.Part] = [
        generative_models.Part.from_uri(
            data_file["uri"], mime_type=data_file["mime_type"]
        )
        for data_file in self.file_info
        if data_file["mime_type"] in ACCEPTED_FILE_TYPES
    ]

    text_part = generative_models.Part.from_text(full_style_prompt)
    contents.append(text_part)

    if len(contents) == 1:
      LOGGER.info(
          "No files were staged, continuing without brand style files or"
          " examples."
      )

    return generative_models.Content(parts=contents, role="user")



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reusable UI components for Copycat."""

import mesop as me

from copycat.ui import event_handlers
from copycat.ui import states
from copycat.ui import styles


@me.content_component
def row(gap: int = styles.DEFAULT_ROW_AND_COLUMN_GAP, **kwargs) -> None:
  """Creates a row of UI elements."""
  with me.box(
      style=me.Style(display="flex", flex_direction="row", gap=gap, **kwargs)
  ):
    me.slot()


@me.content_component
def column(gap: int = styles.DEFAULT_ROW_AND_COLUMN_GAP, **kwargs) -> None:
  """Creates a column of UI elements."""
  with me.box(
      style=me.Style(display="flex", flex_direction="column", gap=gap, **kwargs)
  ):
    me.slot()


@me.content_component
def header_bar(**kwargs) -> None:
  """Creates a header bar."""
  with me.box(
      style=me.Style(
          background=me.theme_var("surface-container"),
          padding=me.Padding.all(10),
          align_items="center",
          display="flex",
          gap=5,
          justify_content="space-between",
          **kwargs,
      )
  ):
    me.slot()


@me.content_component
def header_section() -> None:
  """Adds a section to the header."""
  with me.box(style=me.Style(**styles.HEADER_SECTION_STYLE)):
    me.slot()


@me.content_component
def conditional_tooltip(
    disabled: bool,
    disabled_tooltip: str = "",
    enabled_tooltip: str = "",
    **kwargs,
) -> None:
  """Adds a tooltip to a UI element depending on whether it is disabled."""
  if disabled and disabled_tooltip:
    with me.tooltip(message=disabled_tooltip, **kwargs):
      me.slot()
  elif not disabled and enabled_tooltip:
    with me.tooltip(message=enabled_tooltip, **kwargs):
      me.slot()
  else:
    me.slot()


@me.content_component
def rounded_box_section(title: str = "", **kwargs) -> None:
  """Adds a rounded box section with an optional title."""
  with me.box(style=me.Style(**(styles.ROUNDED_BOX_SECTION_STYLE | kwargs))):
    if title:
      me.text(title, type=styles.ROUNDED_BOX_SECTION_HEADER_TYPE)
    me.slot()


@me.content_component
def dialog(is_open: bool) -> None:
  """Renders a dialog component.

  The dialog is a full screen overlay that can be used to show additional
  information to the user.

  Args:
    is_open: Whether the dialog is visible or not.
  """
  with me.box(
      style=me.Style(
          background="rgba(0, 0, 0, 0.4)"
          if me.theme_brightness() == "light"
          else "rgba(255, 255, 255, 0.4)",
          display="block" if is_open else "none",
          height="100%",
          overflow_x="auto",
          overflow_y="auto",
          position="fixed",
          width="100%",
          z_index=1000,
      )
  ):
    with me.box(
        style=me.Style(
            align_items="center",
            display="grid",
            height="100vh",
            justify_items="center",
        )
    ):
      with me.box(
          style=me.Style(
              background=me.theme_var("surface-container-lowest"),
              border_radius=20,
              box_sizing="content-box",
              box_shadow=(
                  "0 3px 1px -2px #0003, 0 2px 2px #00000024, 0 1px 5px"
                  " #0000001f"
              ),
              margin=me.Margin.symmetric(vertical="0", horizontal="auto"),
              padding=me.Padding.all(20),
          )
      ):
        me.slot()


@me.component
def snackbar(
    *,
    snackbar_is_visible_name: str,
    label: str,
    action_label: str | None = None,
):
  """Creates a snackbar in the center of the screen.

  Args:
    snackbar_is_visible_name: The name of the app state parameter that controls
      the visibility of the snackbar.
    label: Message for the snackbar
    action_label: Optional message for the action of the snackbar
  """
  is_visible = getattr(me.state(states.AppState), snackbar_is_visible_name)
  with me.box(
      style=me.Style(
          display="block" if is_visible else "none",
          height="100%",
          overflow_x="auto",
          overflow_y="auto",
          position="fixed",
          pointer_events="none",
          width="100%",
          z_index=1000,
      )
  ):
    with me.box(
        style=me.Style(
            align_items="end",
            height="100%",
            display="flex",
            justify_content="center",
        )
    ):
      with me.box(
          style=me.Style(
              align_items="center",
              background=me.theme_var("on-surface-variant"),
              border_radius=5,
              box_shadow=(
                  "0 3px 1px -2px #0003, 0 2px 2px #00000024, 0 1px 5px"
                  " #0000001f"
              ),
              display="flex",
              font_size=14,
              justify_content="space-between",
              margin=me.Margin.all(10),
              padding=me.Padding(top=5, bottom=5, right=5, left=15)
              if action_label
              else me.Padding.all(15),
              pointer_events="auto",
              width=300,
          )
      ):
        me.text(
            label,
            style=me.Style(color=me.theme_var("surface-container-lowest")),
        )
        if action_label:
          me.button(
              action_label,
              on_click=event_handlers.on_click_snackbar_close,
              key=snackbar_is_visible_name,
              style=me.Style(color=me.theme_var("primary-container")),
          )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains the objects that contain google ad copy."""

import re
import pydantic


# See https://support.google.com/google-ads/answer/6371157
def parse_google_ads_special_variables(text: str) -> str:
  """Replaces any special variables used by Google Ads with the default string.

  First replaces dynamic keyword insertion with the default keyword. How the
  text is inserted depends on the capitalization:
    "Buy {KeyWord:my keyword} now" -> "Buy My Keyword now"
    "Buy {Keyword:my keyword} now" -> "Buy My keyword now"
    "Buy {keyword:my keyword} now" -> "Buy my keyword now"
    "Buy {KEYWord:my keyword} now" -> "Buy MY Keyword now"
    "Buy {KeyWORD:my keyword} now" -> "Buy My KEYWORD now"

  Then replaces any other customizers with the default value:
    "Buy {CUSTOMIZER.product:my product} now" -> "Buy my product now"

  Args:
    text: The text to parse.

  Returns:
    The text with the default keyword inserted.
  """
  # Title Case
  pattern = r"\{KeyWord:([^}]+)\}"  # The regex pattern
  text = re.sub(pattern, lambda m: m.group(1).title(), text)

  # First Word Capitalized
  pattern = r"\{Keyword:([^}]+)\}"  # The regex pattern
  text = re.sub(pattern, lambda m: m.group(1).capitalize(), text)

  # Lower Case
  pattern = r"\{keyword:([^}]+)\}"  # The regex pattern
  text = re.sub(pattern, lambda m: m.group(1).lower(), text)

  # Caps First Word
  pattern = r"\{KEYWord:([^}]+)\}"  # The regex pattern
  text = re.sub(
      pattern,
      lambda m: m.group(1).split()[0].upper()
      + " "
      + " ".join(m.group(1).split()[1:]).title(),
      text,
  )

  # Caps Last Word
  pattern = r"\{KeyWORD:([^}]+)\}"  # The regex pattern
  text = re.sub(
      pattern,
      lambda m: " ".join(m.group(1).split()[:-1]).title()
      + " "
      + m.group(1).split()[-1].upper(),
      text,
  )

  # Customizers
  pattern = r"\{CUSTOMIZER\.([a-zA-Z0-9_]+):([^}]+)\}"
  text = re.sub(pattern, lambda m: m.group(2), text)

  return text


class GoogleAd(pydantic.BaseModel):
  """Google ad copy.

  The google ad is defined by a list of headlines and descriptions. The
  headlines and descriptions are each limited to 30 and 90 characters
  respectively. Google Ads combines the headlines and descriptions to create the
  final ad copy.
  """

  headlines: list[str] = []  # The list of headlines for the ad.
  descriptions: list[str] = []  # The list of descriptions for the ad.

  def __str__(self) -> str:
    """Returns the headlines and descriptions as a string.

    The string is constructed by combining all of the headlines and descriptions
    into a single string. The headlines are separated by a pipe character. The
    descriptions are separated by a space character.
    """
    headline = " | ".join(self.headlines)
    description = " ".join(self.descriptions)
    return f"**{headline}**\n{description}"

  def __hash__(self) -> int:
    return hash(str(self))

  @property
  def headline_count(self) -> int:
    """The number of headlines in the ad copy."""
    return len(self.headlines)

  @property
  def description_count(self) -> int:
    """The number of descriptions in the ad copy."""
    return len(self.descriptions)

  def __add__(self, other: "GoogleAd") -> "GoogleAd":
    """Adds two GoogleAds together."""
    return self.__class__(
        headlines=self.headlines + other.headlines,
        descriptions=self.descriptions + other.descriptions,
    )


class GoogleAdFormat(pydantic.BaseModel):
  """The Google Ads format."""

  name: str
  max_headlines: int
  max_descriptions: int
  min_headlines: int
  min_descriptions: int
  max_headline_length: int
  max_description_length: int


# More info: https://support.google.com/google-ads/answer/7684791
RESPONSIVE_SEARCH_AD_FORMAT = GoogleAdFormat(
    name="responsive_search_ad",
    max_headlines=15,
    max_descriptions=4,
    min_headlines=3,
    min_descriptions=2,
    max_headline_length=30,
    max_description_length=90,
)


# More info: https://support.google.com/google-ads/answer/12437745
TEXT_AD_FORMAT = GoogleAdFormat(
    name="text_ad",
    max_headlines=3,
    max_descriptions=2,
    min_headlines=1,
    min_descriptions=1,
    max_headline_length=30,
    max_description_length=90,
)


def get_google_ad_format(ad_format: str) -> GoogleAdFormat:
  """Returns the requested Google Ad format.

  Args:
    ad_format: The ad format to get the Google Ad format for. Must be one of
      "responsive_search_ad" or "text_ad".
  """
  ad_formats = [RESPONSIVE_SEARCH_AD_FORMAT, TEXT_AD_FORMAT]
  ad_format_lookup = {ad_format.name: ad_format for ad_format in ad_formats}

  if ad_format not in ad_format_lookup.keys():
    raise ValueError(
        f"Invalid ad format: {ad_format}. Must be one of"
        f" {ad_format_lookup.keys()}"
    )

  return ad_format_lookup[ad_format]



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The main entrypoint for the Copycat UI."""

import logging

import mesop as me

from copycat.ui import components
from copycat.ui import event_handlers
from copycat.ui import generate_ads_page
from copycat.ui import new_instance_page
from copycat.ui import prepare_data_page
from copycat.ui import setup_page
from copycat.ui import states
from copycat.ui import style_guide_page
from copycat.ui import styles
from copycat.ui import sub_pages


all_sub_pages = sub_pages.SubPages()
all_sub_pages.add_page(
    setup_page.setup,
    nav_text="Setup",
    nav_icon="settings",
)
all_sub_pages.add_page(
    prepare_data_page.prepare_data,
    nav_text="Prepare Data",
    nav_icon="download",
)
all_sub_pages.add_page(
    new_instance_page.new_instance,
    nav_text="New Copycat Instance",
    nav_icon="smart_toy",
)
all_sub_pages.add_page(
    style_guide_page.style_guide, nav_text="Style Guide", nav_icon="edit_note"
)
all_sub_pages.add_page(
    generate_ads_page.generate_new_ads,
    nav_text="Generate New Ads",
    nav_icon="bolt",
)


def starting_dialog(state: states.AppState):
  """The dialog that is shown when the user first opens the UI.

  This is also shown when they want to load a new Google Sheet.

  Args:
    state: The AppState.
  """
  with components.dialog(is_open=state.show_starting_dialog):
    if state.google_sheet_url:
      # Can close if there is already a google sheet URL.
      with me.content_button(
          type="icon",
          on_click=event_handlers.close_starting_dialog,
      ):
        me.icon("close", style=me.Style(color=me.theme_var("outline-variant")))

    with components.column(align_items="center", width="100%", gap=0):
      me.text("Welcome to Copycat", type="headline-5")
      me.text(
          "Please select a Google Sheet to load, or create a new one. Note:"
          " after entering the name or URL, you may first need to click outside"
          " the input box for the buttons to activate.",
          style=me.Style(width=500),
      )
      me.text(
          "WARNING: The data in the Google Sheet you use can be edited by"
          " Copycat, so if you have your data in another google sheet, it's"
          " best to select 'Create New Sheet' here and then copy the data to"
          " the new sheet. Only load an existing sheet if it was a previous"
          " Copycat sheet that you want to continue working on.",
          style=me.Style(
              margin=me.Margin.all(15),
              color=me.theme_var("error"),
              width=400,
              border=me.Border.all(
                  me.BorderSide(
                      width=1, color=me.theme_var("error"), style="solid"
                  )
              ),
              text_align="center",
          ),
          type="body-2",
      )
      with components.row(width="100%", margin=me.Margin(top=15), gap=0):
        with components.column(
            align_items="center",
            width="50%",
            gap=0,
            border=me.Border(right=styles.DEFAULT_BORDER_STYLE),
        ):
          me.text("Create New Sheet", type="headline-6")
          me.input(
              label="Google Sheet Name",
              key="new_google_sheet_name",
              on_blur=event_handlers.update_app_state_parameter,
              value=state.new_google_sheet_name,
              appearance="outline",
              style=me.Style(
                  width="100%",
                  padding=me.Padding.all(0),
              ),
          )
          with components.row(align_items="center"):
            me.button(
                "New",
                type="flat",
                disabled=not state.new_google_sheet_name,
                on_click=event_handlers.create_new_google_sheet,
            )
            me.checkbox(
                label="Include Demo Data",
                key="new_google_sheet_include_demo_data",
                on_change=event_handlers.update_app_state_parameter_checkbox,
                checked=state.new_google_sheet_include_demo_data,
                disabled=not state.new_google_sheet_name,
            )
        with components.column(align_items="center", width="50%", gap=0):
          me.text("Load Existing Sheet", type="headline-6")
          me.input(
              label="Google Sheet URL",
              key="new_google_sheet_url",
              on_blur=event_handlers.update_app_state_parameter,
              value=state.new_google_sheet_url,
              type="url",
              appearance="outline",
              style=me.Style(
                  width="100%",
                  padding=me.Padding.all(0),
              ),
          )
          me.button(
              "Load",
              type="flat",
              disabled=not state.new_google_sheet_url,
              on_click=event_handlers.load_existing_google_sheet,
          )


def main_copycat_header(state: states.AppState):
  """The header bar that is shown at the top of the UI.

  It contains the Copycat name, the Google Sheet URL, buttons to save and
  load the parameters from the Google Sheet, and a drop down to select the
  logging level.

  Args:
    state: The AppState.
  """

  with components.header_bar(
      border=me.Border.symmetric(vertical=styles.DEFAULT_BORDER_STYLE)
  ):
    with components.header_section():
      me.text(
          "Copycat",
          type="headline-3",
          style=me.Style(margin=me.Margin(bottom=0)),
      )

    with components.header_section():
      me.text(
          "Google Sheet URL",
          type="headline-6",
          style=me.Style(margin=me.Margin(bottom=0)),
      )
      me.input(
          label="URL",
          value=state.google_sheet_url,
          type="url",
          appearance="outline",
          style=me.Style(
              width=500,
              padding=me.Padding.all(0),
              margin=me.Margin(top=20),
          ),
          readonly=True,
      )
      with me.content_button(
          style=me.Style(
              padding=me.Padding.symmetric(vertical=30, horizontal=25)
          ),
          on_click=event_handlers.save_params_to_google_sheet,
      ):
        me.icon("save")
        me.text("Save")

      with me.content_button(
          style=me.Style(
              padding=me.Padding.symmetric(vertical=30, horizontal=25)
          ),
          on_click=event_handlers.open_starting_dialog,
      ):
        me.icon("add_circle")
        me.text("New / Load")

      with me.content_button(
          style=me.Style(
              padding=me.Padding.symmetric(vertical=30, horizontal=25)
          ),
          on_click=event_handlers.show_hide_google_sheet,
      ):
        if state.display_google_sheet:
          me.icon("visibility_off")
          me.text("Hide Preview")
        else:
          me.icon("visibility")
          me.text("Show Preview")

    with components.header_section():
      me.select(
          label="Log Level",
          options=[
              me.SelectOption(label="DEBUG", value=str(logging.DEBUG)),
              me.SelectOption(label="INFO", value=str(logging.INFO)),
              me.SelectOption(label="ERROR", value=str(logging.ERROR)),
              me.SelectOption(label="CRITICAL", value=str(logging.CRITICAL)),
          ],
          on_selection_change=event_handlers.update_log_level,
          multiple=False,
          value=str(state.log_level),
          style=me.Style(
              padding=me.Padding.all(0),
              margin=me.Margin(top=20),
          ),
      )


def body_and_google_sheet_preview(state: states.AppState):
  """The main body of the UI.

  It contains the sub pages (left), and the Google Sheet preview if it is
  enabled (right).

  Args:
    state: The AppState.
  """

  with components.row(
      gap=0,
      height="100%",
      width="100%",
  ):
    all_sub_pages.render(
        height="100%",
        width="50%" if state.display_google_sheet else "100%",
        gap=0,
        border=me.Border(right=styles.DEFAULT_BORDER_STYLE),
    )

    # Google Sheet
    if state.display_google_sheet:
      with me.box(
          style=me.Style(
              height="100%",
              width="50%",
          )
      ):
        me.embed(
            src=state.google_sheet_url,
            style=me.Style(width="100%", height="100%"),
        )


@me.page(path="/")
def home():
  """The home page of the Copycat UI.

  This contains all the scaffolding for the UI, but the actual content is
  rendered by the sub-pages.
  """
  state = me.state(states.AppState)

  starting_dialog(state)
  components.snackbar(
      snackbar_is_visible_name="show_copycat_instance_created_snackbar",
      label=(
          "Copycat instance successfully created. You can now generate a style"
          " guide and new ads."
      ),
      action_label="Okay",
  )
  components.snackbar(
      snackbar_is_visible_name="show_ad_copy_generated_snackbar",
      label=(
          "Ad copy generation complete. You can view the generated ads in the"
          " Google Sheet in the 'Generated Ads' tab."
      ),
      action_label="Okay",
  )

  with components.column(height="100%", width="100%", gap=0):
    main_copycat_header(state)
    body_and_google_sheet_preview(state)
